/**
 * Copyright (c) 2020 Guido Kramann
 *
 * This file is part of a library called AOGdogmaWEB - http://www.kramann.info/AOGdogmaWEB.html
 *
 * AOGdogmaWEB is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * AOGdogmaWEB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the AOGdogmaWEB. If not, see <http://www.gnu.org/licenses/>.
*/

function AOGscript()
{

    function makeArray(zeilen, spalten)
    {
         var arr = new Array(zeilen);
         for(var i=0;i<arr.length;i++)
         {
               arr[i] = new Array(spalten);
         }
         return arr;
    }

    function makeArrayZerosZeros(zeilen, spalten)
    {
         var arr = new Array(zeilen);
         for(var i=0;i<arr.length;i++)
         {
               arr[i] = new Array(spalten);
         }
         for(var i=0;i<arr.length;i++)
            for(var k=0;k<arr[i].length;k++)
                arr[i][k] = 0;


         return arr;
    }

    function makeZeros(zeilen)
    {
         var arr = new Array(zeilen);
         for(var i=0;i<arr.length;i++)
         {
               arr[i] = 0;
         }
         return arr;
    }

    var successor;
  
    var operations = "+-*/%~|>";
  
  
    var sourceT; // t_ms, dt_ms, t_ticks

    var MIDIOUT;
    var script;
    var scriptpre;

    //    a = Array(5).fill(0).map(x => Array(10).fill(0))
    var Var; // = makeArray(256, 256); //alle möglichen Variablennamen, die sich mit zwei ASCII-Zeichen bilden lassen
    
    var instName, pulseName;
    var instparam;
    var pulseparam;
    var   t_alt;

    var VERBOSE = false;

    var potencies;
    const prime = [2,3,5,7];

    this.init = function(inMAXVOICES,inMAXLINES,inMAXPULSE,inMAXPOTENCY,inVERBOSE)
    {
        Var = makeArrayZerosZeros(256, 256);

        //potencies = new int[4][MAXPOTENCY];
        potencies = makeArrayZerosZeros(4,inMAXPOTENCY);
        //precalculate potencies:
        for(var i=0;i<prime.length;i++)
        {
             var p = 1;
             for(var k=0;k<potencies[i].length;k++)
             {
                  potencies[i][k] = p;
                  p*=prime[i];
             }
        }
        
      
        VERBOSE = inVERBOSE;
        
        MIDIOUT = makeZeros(inMAXVOICES); //new Array(inMAXVOICES); //new int[MAXVOICES];
        script = new Array(inMAXLINES); //new String[MAXLINES];
        scriptpre = new Array(inMAXLINES); //new String[MAXLINES];
        instName = new Array(inMAXVOICES); //new String[MAXVOICES];
        pulseName = new Array(inMAXPULSE); //new String[MAXPULSE];
        instparam = makeArrayZerosZeros(inMAXVOICES,9); //new int[MAXVOICES][9];
        pulseparam = makeArrayZerosZeros(inMAXPULSE,2); //new int[MAXPULSE][2];
        t_alt = makeZeros(inMAXPULSE); //new Array(inMAXPULSE); //new int[MAXPULSE];
    }
  
    function setVar(inname,value)
    {
         if(inname==null) {return;}

         var name = inname.trim();

         if(name.length==0) {return;}
        

         if(name.length==1)
         {
               //var[(int)name.charAt(0)][0] = value;  //   charCodeAt(0)
               Var[name.charCodeAt(0)][0] = value;
         }
         else if(name.length==2)
         {
               //var[(int)name.charAt(0)][(int)name.charAt(1)] = value;
               Var[name.charCodeAt(0)][name.charCodeAt(1)] = value;
         }
    }
  
    function getVar(inname)
    {
         if(inname==null) {return;}

         var name = inname.trim();

         if(name.length==1)
         {
               //return var[(int)name.charAt(0)][0];
               return Var[name.charCodeAt(0)][0];
         }
         else if(name.length==2)
         {
               //return var[(int)name.charAt(0)][(int)name.charAt(1)];
               return Var[name.charCodeAt(0)][name.charCodeAt(1)];
         }
         else
         {
               return -1;
         }
    }

     function frequency2Midi(freq)
     {
        if(freq<=1.0)
        {
             return 0.0; //Nullereignisse behandeln
        }

        var freq69 = Math.pow(2.0,69.0/12.0);
        return 12.0*Math.log(freq*freq69/440.0)/Math.log(2.0);
     }

//var TT=0;  
    this.step = function()
    {
//console.log("xSTEP"+TT);
//TT++;
        if(script==null)
        {
              return;
        }

        if(script.length==0)
        {
              return;
        }

        var ZEILANZ = 0;
        for(var i=0;i<script.length;i++)
        {
             if(scriptpre[i]!=null)
             {
                  script[i] = scriptpre[i];
                  ZEILANZ++;
             }
             else
             {
                  script[i] = null;
             }
        } 
        

        if(VERBOSE)
        {
            console.log("step tick="+sourceT[2]);
        }
        //println("step tick="+sourceT[2]);
        

        var anzInst = 0;
        var anzPulse = 0;
        var anzFormula = 0;
        
        //var t_ms = (int)sourceT[0]; 
        var t_ms = sourceT[0]; 
        
        for(var i=0;i<ZEILANZ;i++)
        {
            var s=script[i];
            if(s!=null)
            {
               s = s.trim();
            }

            if(s!=null && s.length==0) //Leerzeile
            {
            }
            else if(s!=null && s.charAt(0)=='#') //Zeilenkommentar
            {
            }
            else if(s!=null && s.includes("!")) //pulse
            {
                pulseName[anzPulse] = getName(s);                
                
                //extractNumbers(pulseparam[anzPulse],s);
                //statt dessen auch Variablen zulassen:
                var inx0 = 0;
                var inx1 = nextWhitespace(inx0,s); 

                pulseparam[anzPulse][0] = extractValue(s.substring(inx0,inx1));
                inx0 = inx1+1;
                inx1 = nextWhitespaceOrChar(inx0,s,'!'); 
                pulseparam[anzPulse][1] = extractValue(s.substring(inx0,inx1));
                                
//if(i==0)
//document.getElementById("test").value = "pulseparam[anzPulse][0]="+pulseparam[anzPulse][0]+" pulseparam[anzPulse][1]="+pulseparam[anzPulse][1]
//+" pulseName[anzPulse]="+pulseName[anzPulse];
//+result+" inx0="+inx0+" inx1="+inx1+" name="+s.substring(inx0,inx1)+" extractValue="+extractValue(s.substring(inx0,inx1));
//+result+" inx0="+inx0+" inx1="+inx1+" name="+s.substring(inx0,inx1)+" getVar()="+getVar(s.substring(inx0,inx1).trim());
                
                //Aktuellen Wert berechnen und speichern:
                //                t0            + t    / dt
                //int t = pulseparam[anzPulse][0] + t_ms / pulseparam[anzPulse][1];
        
                //Alternative, die ev. weniger "klappert": bei dt=10ms
                var dt_ms = Math.floor(sourceT[1]);
                var ticks = Math.floor(sourceT[2]);
                var t = Math.floor(   pulseparam[anzPulse][0] + (ticks * dt_ms) / pulseparam[anzPulse][1]   );
                
        
                //Nur ungleich Null, wenn > alter Wert
                if(t>t_alt[anzPulse])
                {
                     t_alt[anzPulse] = t;
                     setVar(pulseName[anzPulse],t);
//if(i==0)
//document.getElementById("test").value = " getVar(pulseName[anzPulse])="+getVar(pulseName[anzPulse]);
                }
                else
                {
                     // ... sonst inaktiv in diesem Zyklus!
                     setVar(pulseName[anzPulse],0);
                }
                
                anzPulse++;
            }
            else if(s!=null && s.includes(":")) //instrument
            {
                instName[anzInst] = getName(s);
                
                //extractNumbers(instparam[anzInst],s);
                //auch Variablen zulassen:
                var inx0 = 0;
                var inx1 = nextWhitespace(inx0,s); //sucht nächstes Leezeichen
                instparam[anzInst][0] = extractValue(s.substring(inx0,inx1));
                for(var ii=1;ii<8;ii++)
                {
                    inx0 = inx1+1;
                    inx1 = nextWhitespace(inx0,s); //sucht nächstes Leezeichen
                    instparam[anzInst][ii] = extractValue(s.substring(inx0,inx1));
                }
                
                inx0 = inx1+1;
                inx1 = nextWhitespaceOrChar(inx0,s,':'); //sucht nächstes Leezeichen
                instparam[anzInst][8] = extractValue(s.substring(inx0,inx1));
                
                //Null eintragen, Default, wenn Formeln nichts liefern
                setVar(instName[anzInst],0);
                //entspricht auch MIDIOUT in natürlicher Reihenfolge:
                MIDIOUT[anzInst] = 0;
                
                anzInst++;
            }
            //Vor Auftreten der ersten Formel müssen bereits alle Pulse und Instrumente definiert sein
            //Formeln können ihr Ergebnis einfach gewöhnlichen Variablen zuweisen.
            //Das passiert, wenn der Name kein Instrumentenname ist.
            //Argumente von Formeln dürfen auch zuvor definierte normale Variablen sein, aber keine Instrumente
            // wenn gar keine Zuweisung mit > erfolgt, kann die Zeile ignoriert werden
            
            //  Es ist egal welches Tempo der Grundpuls hat, man kann ja die Millisekunden als Grundlage für die Berechnung der Zustände der abgeleiteten Pulse nehmen,
            //  ist aber ev. ein gewisser Aufwand.
            
            else if(s!=null && s.includes("~")) //formula
            {
                if(VERBOSE) 
                {
                    console.log("++ Formel auswerten ++");
                }

                //Erster Eintrag in Formel muß ein Wert sein:
                var inx0 = 0;
                var inx1 = nextOperation(inx0,s); //sucht nächste Operation

                var value = extractValue(s.substring(inx0,inx1));
                var result = value; //Zwischenergebnis der Folge der Operationen
                var ABBRUCH = false;

                if(result==0)
                {
                   ABBRUCH = true;
                }  

                if(VERBOSE) 
                {
                     console.log("Gefundener Eingangswert t="+value+" exztrahiert aus "+s.substring(inx0,inx1));   
                }  
//                while(!ABBRUCH && (inx1<s.length() || inx0<inx1))
//if(i==1)
//document.getElementById("test").value = "~ gefunden in "+s+" s.length="+s.length+"\n result="
//+result+" inx0="+inx0+" inx1="+inx1+" name="+s.substring(inx0,inx1)+" getVar()="+getVar(s.substring(inx0,inx1).trim());
                while(!ABBRUCH && (inx1<s.length))
                {
                    inx0 = inx1;
                    inx1 = nextValue(inx0,s); //sucht nächste Operation
                    var op = s.substring(inx0,inx1);
                    op = op.trim();
                    var opc = op.charAt(0);
                    inx0 = inx1;
                    inx1 = nextOperation(inx0,s); //sucht nächste Operation
                    if(opc=='~')
                    {
                         var name = s.substring(inx0,inx1);
                         name = name.trim();
                         
                         if(VERBOSE) console.log("## Auskopplung wg. ~ ## name = \""+name+"\" s="+s+" inx0="+inx0+" inx1="+inx1);
//document.getElementById("test").value = "MIDI: "+MIDI[0]+" "+MIDI[1]+" "+MIDI[2]+" ";
//document.getElementById("test").value = "script.step() ## Auskopplung wg. ~ ## name = \""+name+"\" s="+s+" inx0="+inx0+" inx1="+inx1;
                      
                         //Unterscheiden, ob Instrument oder gew. Variable:
                         var INR = isInstrument(name,anzInst);
                         //if(result>0)  //egal, ABBRUCH wird ja geprüft, besser, so Nulldurchgänge z.B. bei Minus möglich
                         //{
                             if(INR>=0)
                             {
                                 //Wert herauslösen und auf Instrument übertragen:
                                 //                         2^p            *             3^q               *          5^r                  *          7^s                   * REST                             
                                 var BASE = potencies[0][instparam[INR][0]]*potencies[1][instparam[INR][1]]*potencies[2][instparam[INR][2]]*potencies[3][instparam[INR][3]] * instparam[INR][4];
                                 var BASE_START = BASE;
                                 var FACTOR = instparam[INR][5]/100.0;
                                 var MIDISHIFT = instparam[INR][6];
                                 var midimin = instparam[INR][7];
                                 var midimax = instparam[INR][8];
                                 
                                 var x = result;
                                 BASE=Math.round(BASE); x=Math.round(x);
                                 while(BASE>=2 && BASE%2==0 && x>=2 && x%2==0) {BASE/=2; x/=2; BASE=Math.round(BASE); x=Math.round(x);}
                                 while(BASE>=3 && BASE%3==0 && x>=3 && x%3==0) {BASE/=3; x/=3; BASE=Math.round(BASE); x=Math.round(x);}
                                 while(BASE>=5 && BASE%5==0 && x>=5 && x%5==0) {BASE/=5; x/=5; BASE=Math.round(BASE); x=Math.round(x);}
                                 while(BASE>=7 && BASE%7==0 && x>=7 && x%7==0) {BASE/=7; x/=7; BASE=Math.round(BASE); x=Math.round(x);}

                                 var midi = 0;
                                 if(BASE>0 && BASE<BASE_START)
                                 {
                                      var f = BASE * FACTOR;
                                      //midi = MIDISHIFT + Math.round(frequency2Midi(f));
                                      //NEW:
                                      midi = MIDISHIFT - 24 + Math.round(frequency2Midi(f));
                                      if(midi>=midimin && midi<=midimax)
                                      {
                                            MIDIOUT[INR] = midi;
                                            setVar(name,midi);                                            
                                      }
                                 }
                                 
                                 if(VERBOSE) console.log("Auskopplung auf Instrument Nr."+INR+" name="+name+" midi = "+midi+" getVar="+getVar(name));
                             }                              
                             else
                             {
                                 //Zuweisung zu normaler Variable
                                 setVar(name,result);
                                 
                                 if(VERBOSE) console.log("Auskopplung auf normale Variable name="+name+" result="+result+" getVar="+getVar(name));
                             }
                         //}  // >0     
                    }
                    else
                    {
                        if(VERBOSE) console.log("!! neue Operation, input="+result+ " !!");
                      
                        var operand = extractValue(s.substring(inx0,inx1));
                        //Neues result mit neuer Operation bestimmen
                        //ABBRUCH, wenn Operation nicht möglich:
                        if(opc=='+')
                        {
                           result+=operand;
                           if(VERBOSE) console.log("OP + "+operand);
                        }   
                        else if(opc=='-')
                        {
                           result-=operand;
                           if(VERBOSE) console.log("OP - "+operand);
                        }   
                        else if(opc=='*')
                        {
                           result*=operand;
                           if(VERBOSE) console.log("OP * "+operand);
                           
                        }   
                        else if(opc=='/')
                        {
                           if(operand!=0)
                           {
                               result/=operand;
                               result = Math.floor(result);
                           }
                           else
                           {
                               ABBRUCH = true;
                           }
    
                           if(VERBOSE) console.log("OP / "+operand);

                        }   
                        else if(opc=='%')
                        {
                           if(operand>0)
                               result%=operand;
                           else
                               ABBRUCH = true;
                           if(VERBOSE) console.log("OP % "+operand);
                        }
                        else if(opc=='|')
                        {
                           if(operand>0  && result>=operand && result%operand==0)
                               result=result;
                           else
                               result = 0;
                           if(VERBOSE) console.log("OP | "+operand);
                        }
                        
                        if(VERBOSE) console.log("Neuer Wert: "+result);
                    }    
                }
                anzFormula++;
            }
        }
        
        if(VERBOSE)
        {
           printInfo(anzPulse,anzInst,anzFormula);
        }

        if(successor!=null)
        {
            successor.step();
        }
    }

    function isInstrument(name,anzInst)
    {
         for(var i=0;i<anzInst;i++)
         {
              if(name==instName[i])
              {
                 return i;
              }
         }
         return -1;
    }

    function extractValue(intext)
    {
         if(intext==null)
         {
              return 0;
         }

         var text = intext.trim();

         if(text.length==0)
         {
              return 0;
         }

         //Zahl, wenn Beginn mit Ziffer:
         //  charCodeAt(0)
         //if((int)text.charAt(0)>=(int)'0' && (int)text.charAt(0)<=(int)'9')
         if(text=="0")
         {
              return 0;
         }
         else if(text.charCodeAt(0)>="0".charCodeAt(0) && text.charCodeAt(0)<="9".charCodeAt(0))
         {
              var x = text.charCodeAt(0) - "0".charCodeAt(0);
              for(var i=1;i<text.length;i++)
              {
                   if(text.charCodeAt(i)>="0".charCodeAt(0) && text.charCodeAt(i)<="9".charCodeAt(0))
                   {
                        x*=10;
                        x+=text.charCodeAt(i) - "0".charCodeAt(0);
                   }
                   else
                   {
                        break;
                   }
              }
              return x;
         }
         else
         {
              return getVar(text); 
         }
    }

    function isOperation(c)
    {
          for(var i=0;i<operations.length;i++)
              if(operations.charAt(i)==c)
                  return true;
          return false;        
    }

    function isWhitespace(x)
    {
        if(x==null)
        {
            return false;
        }

        var s=x.trim();

        if(s.trim()=="")
            return true;
        else
            return false;
    }

    function nextWhitespace(inx,text)
    {
         while( inx<text.length && !isWhitespace(text.charAt(inx)) )
           inx++;

         //Korrektur: Whitespaces überspringen:
         //NACHTRAG IN WEB-ANWENDUNG!!!!
         while( inx+1<text.length && isWhitespace(text.charAt(inx+1)) )
           inx++;
           


         return inx;    
    }
    function nextWhitespaceOrChar(inx,text,character)
    {
         while( inx<text.length && !isWhitespace(text.charAt(inx)) && text.charAt(inx)!=character )
             inx++;

         //Korrektur: Whitespaces überspringen:
         //NACHTRAG IN WEB-ANWENDUNG!!!!
         while( inx+1<text.length && isWhitespace(text.charAt(inx+1)) )
           inx++;
             


         return inx;    
    }
    function nextOperation(inx,text)
    {
         while( inx<text.length && !isOperation(text.charAt(inx)) )
             inx++;
         return inx;    
    }
    function nextValue(inx,text)
    {
         while( inx<text.length && (  isWhitespace(text.charAt(inx)) || isOperation(text.charAt(inx))) )
             inx++;
         return inx;    
    }

    function printInfo(anzPulse, anzInst, anzFormula)
    {
         console.log("\n");
         console.log("*** SCRIPT ***");
         console.log("anzPulse="+anzPulse);
         for(var i=0;i<anzPulse;i++)
         {
              console.log("name="+pulseName[i]+"   ");
              for(var k=0;k<pulseparam[i].length;k++)
                  console.log(pulseparam[i][k]+" ");
              console.log("\n");    
         }
         console.log("anzInst="+anzInst);
         for(var i=0;i<anzPulse;i++)
         {
              console.log("name="+instName[i]+"   ");
              for(var k=0;k<instparam[i].length;k++)
                  console.log(instparam[i][k]+" ");
              console.log("\n");    
         }
         console.log("anzFormula="+anzFormula);
    }

    function getName(text)
    {
         if(isWhitespace( text.charAt(text.length-2) ))
             return text.substring(text.length-1,text.length);  // 1 character
         else    
             return text.substring(text.length-2,text.length);  // 2 characters
    }

    function extractNumbers(num, intext)
    {
          if(num==null) {return;}
          if(intext==null) {return;}

          var text = intext.trim();
 
          if(text.length==0) {return;}


          var anz = 0;
          for(var i=0;i<text.length;i++)
          {
                var x = text.charCodeAt(i);
                if(x>="0".charCodeAt(0) && x<="9".charCodeAt(0))
                {
                     var y = x-"0".charCodeAt(0);
                     for(var k=i+1;k<text.length;k++)
                     {
                          if(i+1<text.length && text.charCodeAt(i+1)>="0".charCodeAt(0) && text.charCodeAt(i+1)<="9".charCodeAt(0))
                          {
                               y*=10;
                               y+=text.charCodeAt(i+1) - "0".charCodeAt(0);
                               i++;
                          }
                          else
                          {
                               break;
                          }
                     }
                     if(anz<num.length)
                     {
                        num[anz] = y;
                     }
                     anz++;   
                }
                while( i+1<text.length && isWhitespace(text.charAt(i+1)) )
                {
                   i++;
                }
          }
    }
    
    this.getSink = function()
    {
        return MIDIOUT;
    }
    
    this.setSource = function(sourceTin)
    {
        sourceT = sourceTin;
    }
    
    this.registerSuccessor = function(insuccessor)    
    {
         successor = insuccessor;
    }
    
    this.updateScript = function(sneu)
    {
         var ii=0;
         if(sneu!=null && sneu.length>0)
         {
              for(var i=0;i<sneu.length;i++)
              {
                 if(sneu[i]!=null && sneu[i].trim().length>0)
                 {
                    scriptpre[ii] = sneu[i].trim();
                    ii++;
                 }
//console.log(i+".: "+scriptpre[i]);
              }
              for(var i=ii;i<scriptpre.length;i++)
              {
                    scriptpre[i] = null;
              }
         }
    }
}

