/**
 * Copyright (c) 2020 Guido Kramann
 *
 * This file is part of a library called AOGdogmaWEB - http://www.kramann.info/AOGdogmaWEB.html
 *
 * AOGdogmaWEB is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * AOGdogmaWEB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the AOGdogmaWEB. If not, see <http://www.gnu.org/licenses/>.
*/

function Instrument()
{
    var GAINLEFT = 1.0;  
    var GAINRIGHT = 1.0;  

    var SAMP = null;
    var TON  = null;
    var DT = new Array(128);
    var BESTINDEX = new Array(128);

    var LR = new Array(0,0);
  
    function iabs(x)
    {
      if(x<0)
         return -x;
      else
         return x;
    }

    function midi2Frequency(midi)
    {
        if(midi<1.0)
            return 0.0; //Nullereignisse behandeln
  
        var freq69 = Math.pow(2.0,69.0/12.0);
        return Math.pow(2.0,midi/12.0)*440.0/freq69;
    }

    this.init = function(inSAMP,inTON,GL,GR)
    {
       GAINLEFT = GL;
       GAINRIGHT = GR;

       //kein this sonst ist diese Funktion gemeint!!!
       SAMP = inSAMP;
       TON  = inTON;
       DT = new Array(128);
       BESTINDEX = new Array(128);

       LR = new Array(0,0);

       findMapping();
    }


    function findMapping()
    {
       for(var i=1;i<DT.length;i++)
       {
         var MIDI = i;
         var DIFFBEST = 500;
         for(var k=0;k<TON.length;k++)
         {
             if(iabs(MIDI - TON[k])<=DIFFBEST)
             {
                 DIFFBEST = iabs(MIDI - TON[k]);
                 BESTINDEX[i] = k;
             }
         }

         var MIDINEXT = TON[ BESTINDEX[i] ];
         //Bestimmung von DT: 
         var FIST = midi2Frequency(MIDINEXT);
         var FSOLL = midi2Frequency(i);
         DT[i] = FSOLL / FIST;
       }
    }

    function berechneSamplewert(sample,position)
    {
        var wert = 0;
        wert+=sample.charCodeAt(position)-41;
        wert*=50;
        wert+=sample.charCodeAt(position+1)-41;
        wert*=50;
        wert+=sample.charCodeAt(position+2)-41;
        wert/=124999; //ergibt wert zwischen 0..1
        wert*=2;
        return wert-1.0;
    }


    this.getDT = function(MIDIAKT)
    {
        return DT[MIDIAKT];
    }

    this.evaluateLR = function(MIDIAKT,POS)
    {
        var INDEX = BESTINDEX[MIDIAKT]; 
        var POSA = Math.floor(POS);
        var POSB = Math.ceil(POS);

        if(INDEX<0 || INDEX>=SAMP.length || POSB*3+3>=SAMP[INDEX].length-2)
        {
             return null;
        }

        var AA = berechneSamplewert(SAMP[INDEX],POSA*3);
        var BB = berechneSamplewert(SAMP[INDEX],POSB*3);

        var FA = 0.5;
        var FB = 0.5;
        if(POSA<POSB)
        {
            FA = POSB - POS;
            FB = POS - POSA;
        }

        LR[0]=AA*FA+BB*FB;
        LR[1]=LR[0];

        LR[0]*=GAINLEFT;
        LR[1]*=GAINRIGHT;

        return LR;
    }
}

