/**
 * Copyright (c) 2020 Guido Kramann
 *
 * This file is part of a library called AOGdogmaWEB - http://www.kramann.info/AOGdogmaWEB.html
 *
 * AOGdogmaWEB is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * AOGdogmaWEB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the AOGdogmaWEB. If not, see <http://www.gnu.org/licenses/>.
*/
function Pulse()
{
    var successor = null; 
    
    var start_time_ms = 0;
    
    /** Control Parameter for Thread: cycle_time_ms */
    var source_data = null;//
    
    /** output: {t_ms, dt_ms, t_ticks} */
    var sink_data = null;  
    
    var schedExecService = null;
    
    this.getSink = function()
    {
         return sink_data;
    }
    
    this.registerSuccessor = function(insuccessor)
    {
           successor = insuccessor;
    }    

    /**
     *  @param  dt_ms of type long[1], contains new period for thread, will be actualized with next call of run(), resp. step() 
     */
    this.setSource = function(dt_ms)
    {
          source_data = dt_ms;
    }

    function currentTimeMillis()
    {
        return (new Date()).getTime();
    }

    function initialize(dt_ms)
    {
        //start_time_ms = System.currentTimeMillis();
        start_time_ms = currentTimeMillis();

        sink_data = new Array(3);
        sink_data[0] = 0; // t in ms
        sink_data[1] = dt_ms; // dt in ms
        sink_data[2] = 0; // t_ticks -- number of calls of step() resp. run()
        
        source_data = new Array(1);
        source_data[0] = dt_ms;
    }
    
    /**
     * @param  dt_ms  time period of one call of run() resp. step() in milliseconds
     * Starts thread.
     */
    this.init =  function(dt_ms)
    {
        initialize(dt_ms);
        //schedExecService = Executors.newSingleThreadScheduledExecutor();
        //schedExecService.scheduleAtFixedRate(this, 0, dt_ms, TimeUnit.MILLISECONDS);        
console.log("Starte stepper");
        schedExecService = setInterval(step, dt_ms); 
    }
    
//var TT=0;
    function step()
    {
//console.log("STEP"+TT);
//TT++;
        //check wether dt_ms has changed:
        if(source_data!=null && source_data[0]!=sink_data[1])
        {
            sink_data[1] = source_data[0];
            //schedExecService.scheduleAtFixedRate(this, 0, sink_data[1], TimeUnit.MILLISECONDS); 
            if(schedExecService!=null)
                schedExecService.clearInterval(prozess); 
            schedExecService = setInterval(step, sink_data[1]); 
        }
        
        if(successor!=null)
        {
            successor.step();          
        }
        sink_data[0] = currentTimeMillis() - start_time_ms; // t in ms
        sink_data[2]++; // ticks            
    }
    
    /**
     * Returns time elapsed since start of thread in ms, granularity according to calls of step() resp. run().
     */
    this.get_t_ms = function()
    {
         return sink_data[0];
    }
    
    /**
     * Returns period of one call of step() resp. run().
     */    
    this.get_dt_ms = function()
    {
         return sink_data[1];
    }
    
    
    /**
     * Returns number of calls of step() resp. run() since start of thread.
     */    
    this.get_t_ticks = function()
    {
         return sink_data[2];
    }
}


